from simple_duck_ml.dataset_unpacker.i_dataset_unpacker import IDatasetUnpacker
from simple_duck_ml.dataset_unpacker.dataset import Dataset
from typing import Callable, Literal, Optional
from numpy.typing import NDArray
from struct import unpack
import numpy as np
import struct
import cv2

class BinDatasetUnpacker(IDatasetUnpacker):
    def __init__(self, path: str) -> None:
        self.path = path
        self.file_buffer = open(path, "rb")

    def unpack(
        self,
        label: int,
        qnt: int | Literal["*"]="*",
        normalization: Optional[Callable[[NDArray], NDArray]]=None,
    ) -> Dataset:
        self.file_buffer.seek(0)

        if isinstance(qnt, str) and qnt != "*":
            raise ValueError("Error: 'qnt' argument should be integer or Literal['*']")
        
        images = []
        while True:
            try:
                img = self.__unpack_one()
                if img is None:
                    break
                
                if normalization is not None:
                    img = normalization(img)
                
                images.append(img)
                if qnt != "*" and len(images) >= qnt:
                    break

            except struct.error:
                break

        if not images:
            raise TypeError(f"Could not unpack any image from {self.file_buffer.name}!")

        # Stack all images into one tensor
        images_tensor = np.stack(images, axis=0)
        return Dataset(images_tensor, label)

    def __unpack_one(self) -> Optional[NDArray[np.uint8]]:
        """
        Unpack a single sketch image from the binary dataset file.
        """
        image = np.ones((256, 256), dtype=np.uint8) * 255

        # Skip header (8 + 2 + 1 + 4 = 15 bytes)
        self.file_buffer.read(8)
        self.file_buffer.read(2)
        self.file_buffer.read(1)
        self.file_buffer.read(4)

        n_strokes_data = self.file_buffer.read(2)
        if len(n_strokes_data) < 2:
            return None

        n_strokes, = unpack("H", n_strokes_data)
        for _ in range(n_strokes):
            n_points_data = self.file_buffer.read(2)
            if len(n_points_data) < 2:
                return None

            n_points, = unpack("H", n_points_data)
            x_data = self.file_buffer.read(n_points)
            y_data = self.file_buffer.read(n_points)

            if len(x_data) < n_points or len(y_data) < n_points:
                return None

            x_vec = struct.unpack(f"{n_points}B", x_data)
            y_vec = struct.unpack(f"{n_points}B", y_data)

            for i in range(len(x_vec) - 1):
                pt1 = (int(x_vec[i]), int(y_vec[i]))
                pt2 = (int(x_vec[i + 1]), int(y_vec[i + 1]))
                cv2.line(image, pt1, pt2, color=(0,), thickness=3)

        return image




from duckdi import register

from simple_duck_ml.activations.relu_activation import ReLuActivation
from simple_duck_ml.activations.softmax_activation import SoftmaxActivation
from simple_duck_ml.dataset_unpacker.bin_dataset_unpacker import BinDatasetUnpacker
from simple_duck_ml.layers.convolutional_layer import ConvolutionalLayer
from simple_duck_ml.layers.dense_layer import DenseLayer
from simple_duck_ml.layers.flatten_layer import FlattenLayer
from simple_duck_ml.loss.cross_entropy_loss import CrossEntropyLoss
from simple_duck_ml.loss.mse_loss import MSELoss

# Activations
register(ReLuActivation, label="relu")
register(SoftmaxActivation, label="softmax")

# Layers
register(ConvolutionalLayer, label="conv")
register(DenseLayer, label="dense")
register(FlattenLayer, label="flat")

# Loss
register(CrossEntropyLoss, label="cross_entropy")
register(MSELoss, label="mse")

# DatasetUnpacker
register(BinDatasetUnpacker, label="bin")



from numpy.typing import NDArray
from typing import Dict
import numpy as np
import os

def write_tensor(tensors: Dict[str, NDArray[np.float64]], path: str, overwrite: bool=True) -> str:
    path = path if path.endswith(".npz") else f"{path}.npz"
    if os.path.isfile(path) and not overwrite:
        raise FileExistsError(f"Error: \"{path}\" already exists\nSet \"overwrite=True\" to overwrite file!")

    np.savez_compressed(file=path, allow_pickle=True, **tensors)
    return path

    
def load_tensor(path: str, find_on_path: bool) -> Dict[str, NDArray[np.float64]]:
    if find_on_path and not path.endswith(".npz") and os.path.isdir(path):
        files = [file for file in os.listdir(path) if file.endswith(".npz")]
        if files:
           path = os.path.join(path, files[0])

    if not os.path.isfile(path):
        raise FileNotFoundError(f"Error: Could not find \".npz\" file on \"{path}\"!")   

    with np.load(path, allow_pickle=True) as data:
        tensors = { key: data[key] for key in data.files }

    return tensors   



from abc import ABC, abstractmethod
from numpy.typing import NDArray
from duckdi import Interface
from typing import Optional, Self
import numpy as np

@Interface(label='layer')
class ILayer(ABC):
    name: str

    @abstractmethod
    def forward(self, x: NDArray[np.float64]) -> NDArray[np.float64]: ...

    @abstractmethod
    def backward(self, delta: NDArray[np.float64]) -> NDArray[np.float64]: ...
    
    @abstractmethod
    def clean_grad(self) -> None: ...

    @abstractmethod
    def update(self, learning_rate: float = 0.01, batch_size: int = 1) -> None: ...

    @classmethod
    def load(cls, path: str) -> Self: ...

    @abstractmethod
    def save(self, name: Optional[str] = None, path: str = ".", overwrite: bool=True) -> str: ...



from typing import Dict
import toml
import os

def write_toml(obj: Dict, path: str, overwrite: bool=True) -> str:
    path = path if path.endswith(".toml") else f"{path}.toml"
    if os.path.isfile(path) and not overwrite:
        raise FileExistsError(f"Error: \"{path}\" already exists\nSet \"overwrite=True\" to overwrite file!")

    with open(path, "w") as file:
        file.write(toml.dumps(obj))

    return path


def load_toml(path: str, find_on_path: bool=False) -> Dict:
    if find_on_path and not path.endswith(".toml") and os.path.isdir(path):
        files = [file for file in os.listdir(path) if file.endswith(".toml")]
        path = files[0] if files else path

    if not os.path.isfile(path):
        raise FileNotFoundError(f"Error: Could not find \".toml\" file on \"{path}\"!")   

    with open(path, "r") as file:
        data = toml.load(file)

    return data



from simple_duck_ml.dataset_unpacker.dataset import Dataset
from typing import Callable, Literal, Optional
from abc import ABC, abstractmethod
from numpy.typing import NDArray
from duckdi import Interface

@Interface(label="dataset_unpacker")
class IDatasetUnpacker(ABC):
    path: str
    
    @abstractmethod
    def unpack(
        self,
        label: int,
        qnt: int | Literal["*"]="*",
        normalization: Optional[Callable[[NDArray], NDArray]]=None,
    ) -> Dataset: ...





from abc import ABC, abstractmethod
from duckdi import Interface
from numpy.typing import NDArray
import numpy as np

@Interface(label='loss')
class ILoss(ABC):
    name: str

    @abstractmethod
    def __call__(self, y_pred: NDArray[np.float64], y_true: NDArray[np.float64]) -> float: ...
    
    @abstractmethod
    def derivative(self, y_pred: NDArray[np.float64], y_true: NDArray[np.float64]) -> NDArray[np.float64]: ...




from simple_duck_ml.loss.i_loss import ILoss
from numpy.typing import NDArray
import numpy as np

class CrossEntropyLoss(ILoss):
    name = 'cross_entropy'

    def __call__(self, y_pred: NDArray[np.float64], y_true: NDArray[np.float64]) -> float:
        if y_true.shape != y_pred.shape:
            y_true = y_true.reshape(y_pred.shape)

        eps = 1e-12
        y_pred = np.clip(y_pred, eps, 1 - eps)

        return -np.mean(np.sum(y_true * np.log(y_pred), axis=-1))

    def derivative(self, y_pred: NDArray[np.float64], y_true: NDArray[np.float64]) -> NDArray[np.float64]:
        return y_pred - y_true




from numpy.typing import NDArray
from simple_duck_ml.activations.i_activation import IActivation
from simple_duck_ml.activations.softmax_activation import SoftmaxActivation
from simple_duck_ml.layers.i_layer import ILayer
from typing import Dict, Optional, Self, Type
import numpy as np

from simple_duck_ml.serializers.tensor_io import load_tensor, write_tensor
from simple_duck_ml.serializers.toml_io import load_toml, write_toml
from duckdi import Get
import uuid
import os

class DenseLayer(ILayer):
    name = "dense"

    def __init__(self, output_size: int, activation: IActivation) -> None:
        self.output_size = output_size
        self.activation = activation

        self.input_size: Optional[int] = None
        self.w: Optional[NDArray[np.float64]] = None
        self.b: Optional[NDArray[np.float64]] = None
        self._x: Optional[NDArray[np.float64]] = None
        self._z: Optional[NDArray[np.float64]] = None
        self._output: Optional[NDArray[np.float64]] = None

        self._grad_w: Optional[NDArray[np.float64]] = None
        self._grad_b: Optional[NDArray[np.float64]] = None


    def _init_params(self, input_size: int) -> None:
        self.input_size = input_size

        limit = np.sqrt(2.0 / self.input_size)
        self.w = np.random.randn(self.output_size, self.input_size) * limit
        self.b = np.zeros((self.output_size, 1))
        self._grad_w = np.zeros_like(self.w)
        self._grad_b = np.zeros_like(self.b)


    def forward(self, x: NDArray[np.float64]) -> NDArray[np.float64]:
        if x.ndim == 1:
            x = x.reshape(-1, 1)

        if self.w is None or self.b is None or self._grad_w is None or self._grad_b is None:
            self._init_params(x.shape[0])

        if self.w is None:
          raise RuntimeError("Error: Could Not Initialize Dense Layer Weights") 

        z = np.dot(self.w, x) + self.b
        
        self._x = x
        self._z = z

        self._output = self.activation(z)
        return self._output


    def backward(self, delta: NDArray[np.float64]) -> NDArray[np.float64]:
        if self._output is None or self._x is None or self._z is None:
            raise RuntimeError("Forward should be called before Backward")

        if self.w is None:
          raise RuntimeError("Error: Could Not Initialize Dense Layer Weights")

        # Ensure (classes, batch) shape
        delta = delta.reshape(self.output_size, -1)

        if not isinstance(self.activation, SoftmaxActivation):
            delta *= self.activation.derivative(self._z)

        batch_size = delta.shape[1]
        self._grad_w += np.dot(delta, self._x.T) / batch_size
        self._grad_b += np.mean(delta, axis=1, keepdims=True)

        d_x = np.dot(self.w.T, delta)
        return d_x


    def update(self, learning_rate: float = 0.01, batch_size: int = 1) -> None:
        if self._grad_w is None or self._grad_b is None:
            return

        if self.w is None or self.b is None:
          raise RuntimeError("Error: Could Not Initialize Dense Layer Weights/Bias") 

        self.w -= learning_rate * (self._grad_w / batch_size)
        self.b -= learning_rate * (self._grad_b / batch_size)

        self.clean_grad()
        self._x = None
        self._output = None
        self._z = None


    def clean_grad(self) -> None:
        if self._grad_w is not None:
            self._grad_w.fill(0)

        if self._grad_b is not None:
           self._grad_b = np.zeros_like(self.b)

    def save(self, name: Optional[str] = None, path: str = ".", overwrite: bool=True) -> str:
        if self.w is None or self.b is None:
            raise RuntimeError("Error: Could Not Instance Dense Layer Weights/Bias") 

        name = str(uuid.uuid4()).replace("-", "") if name is None else name
        os.mkdir(os.path.join(path, name))
        file_path = os.path.join(path, name, name)

        tensors_path = write_tensor(
            tensors={ "w": self.w, "b": self.b },
            path=file_path,
            overwrite=overwrite
        )

        return write_toml(
            obj={
                "output_size": self.output_size,
                "input_size": self.input_size,
                "activation": self.activation.name,
                "layer_type": self.name,
                "tensors_path": tensors_path,
            },
            path=file_path,
            overwrite=overwrite,
        )
    
    @classmethod
    def load(cls, path: str) -> Self:
        layer_info = load_toml(path, find_on_path=True)

        def __process_keys[T](obj: Dict, key: str, expected_type: Type[T]) -> T:
            data = obj.get(key, None)
            if data is None or not isinstance(data, expected_type):
                raise KeyError(f"Error: Could Not Process \"{key}\" key!")

            return data
        
        output_size = __process_keys(layer_info, "output_size", int)
        activation = Get(
            IActivation,
            label='activation',
            adapter=__process_keys(layer_info, "activation", str),
        )
        input_size = __process_keys(layer_info, "input_size", int)
        tensors_path = __process_keys(layer_info, "tensors_path", str)

        tensor_info = load_tensor(tensors_path, find_on_path=True)
        w = __process_keys(tensor_info, "w", np.ndarray)
        b = __process_keys(tensor_info, "b", np.ndarray)
        
        layer = cls(output_size, activation)
        
        layer._init_params(input_size)
        layer.w = w
        layer.b = b

        return layer



from simple_duck_ml.loss.i_loss import ILoss
from numpy.typing import NDArray
import numpy as np

class MSELoss(ILoss):
    name = "mse"

    def __call__(self, y_pred: NDArray[np.float64], y_true: NDArray[np.float64]) -> float:
        return float(np.mean((y_pred - y_true) ** 2))

    def derivative(self, y_pred: NDArray[np.float64], y_true: NDArray[np.float64]) -> NDArray[np.float64]:
        y_true = y_true.reshape(y_pred.shape)
        return 2 * (y_pred - y_true) / y_true.size
    
